package java.util;

import java.util.function.Consumer;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.lang.reflect.Array;
import java.io.Serializable;

public class LinkedList<E> extends AbstractSequentialList<E> implements List<E>, Deque<E>, Cloneable, Serializable
{
    transient int size;
    transient LinkedList.Node<E> first;
    transient LinkedList.Node<E> last;
    private static final long serialVersionUID = 876323262645176354L;
    
    public LinkedList() {
        this.size = 0;
    }
    
    public LinkedList(Collection<? extends E> collection) {
        this();
        this.addAll(collection);
    }
    
    private void linkFirst(E e) {
        LinkedList.Node<Object> first = this.first;
        LinkedList.Node last = new LinkedList.Node(null, e, first);
        this.first = last;
        if (first == null) {
            this.last = last;
        }
        else {
            first.prev = last;
        }
        ++this.size;
        ++this.modCount;
    }
    
    void linkLast(E e) {
        LinkedList.Node<Object> last = this.last;
        LinkedList.Node first = new LinkedList.Node(last, e);
        this.last = first;
        if (last == null) {
            this.first = first;
        }
        else {
            last.next = first;
        }
        ++this.size;
        ++this.modCount;
    }
    
    void linkBefore(E e, LinkedList.Node<E> node) {
        LinkedList.Node<Object> prev = node.prev;
        LinkedList.Node<Object> first = new LinkedList.Node(prev, e, node);
        node.prev = first;
        if (prev == null) {
            this.first = first;
        }
        else {
            prev.next = first;
        }
        ++this.size;
        ++this.modCount;
    }
    
    private E unlinkFirst(LinkedList.Node<E> node) {
        Object item = node.item;
        LinkedList.Node<Object> next = node.next;
        node.item = null;
        node.next = null;
        this.first = next;
        if (next == null) {
            this.last = null;
        }
        else {
            next.prev = null;
        }
        --this.size;
        ++this.modCount;
        return item;
    }
    
    private E unlinkLast(LinkedList.Node<E> node) {
        Object item = node.item;
        LinkedList.Node<Object> prev = node.prev;
        node.item = null;
        node.prev = null;
        this.last = prev;
        if (prev == null) {
            this.first = null;
        }
        else {
            prev.next = null;
        }
        --this.size;
        ++this.modCount;
        return item;
    }
    
    E unlink(LinkedList.Node<E> node) {
        Object item = node.item;
        LinkedList.Node<Object> next = node.next;
        LinkedList.Node<Object> prev = node.prev;
        if (prev == null) {
            this.first = next;
        }
        else {
            prev.next = next;
            node.prev = null;
        }
        if (next == null) {
            this.last = prev;
        }
        else {
            next.prev = prev;
            node.next = null;
        }
        node.item = null;
        --this.size;
        ++this.modCount;
        return item;
    }
    
    @Override
    public E getFirst() {
        LinkedList.Node<Object> first = this.first;
        if (first == null) {
            throw new NoSuchElementException();
        }
        return first.item;
    }
    
    @Override
    public E getLast() {
        LinkedList.Node<Object> last = this.last;
        if (last == null) {
            throw new NoSuchElementException();
        }
        return last.item;
    }
    
    @Override
    public E removeFirst() {
        LinkedList.Node<E> first = this.first;
        if (first == null) {
            throw new NoSuchElementException();
        }
        return this.unlinkFirst(first);
    }
    
    @Override
    public E removeLast() {
        LinkedList.Node<E> last = this.last;
        if (last == null) {
            throw new NoSuchElementException();
        }
        return this.unlinkLast(last);
    }
    
    @Override
    public void addFirst(E e) {
        this.linkFirst(e);
    }
    
    @Override
    public void addLast(E e) {
        this.linkLast(e);
    }
    
    @Override
    public boolean contains(Object o) {
        return this.indexOf(o) != -1;
    }
    
    @Override
    public int size() {
        return this.size;
    }
    
    @Override
    public boolean add(E e) {
        this.linkLast(e);
        return true;
    }
    
    @Override
    public boolean remove(Object o) {
        if (o == null) {
            LinkedList.Node<E> node2 = this.first;
            while (node2 != null) {
                if (node2.item == null) {
                    this.unlink(node2);
                    return true;
                }
                node2 = node2.next;
            }
        }
        else {
            LinkedList.Node<E> node = this.first;
            while (node != null) {
                if (o.equals(node.item)) {
                    this.unlink(node);
                    return true;
                }
                node = node.next;
            }
        }
        return false;
    }
    
    @Override
    public boolean addAll(Collection<? extends E> collection) {
        return this.addAll(this.size, collection);
    }
    
    @Override
    public boolean addAll(int n, Collection<? extends E> collection) {
        this.checkPositionIndex(n);
        Object[] array = collection.toArray();
        int length = array.length;
        if (length == 0) {
            return false;
        }
        LinkedList.Node<E> node;
        LinkedList.Node<Object> node2;
        if (n == this.size) {
            node = null;
            node2 = this.last;
        }
        else {
            node = this.node(n);
            node2 = node.prev;
        }
        Object o2;
        for (int length2 = array.length, i = 0; i < length2; node2 = o2, ++i) {
            Object o = array[i];
            o2 = new LinkedList.Node<Object>(node2, o);
            if (node2 == null) {
                this.first = o2;
            }
            else {
                node2.next = o2;
            }
        }
        if (node == null) {
            this.last = node2;
        }
        else {
            node2.next = node;
            node.prev = node2;
        }
        this.size += length;
        ++this.modCount;
        return true;
    }
    
    @Override
    public void clear() {
        LinkedList.Node<Object> next;
        for (LinkedList.Node first = this.first; first != null; first = next) {
            next = first.next;
            first.item = null;
            first.next = null;
            first.prev = null;
        }
        LinkedList.Node<E> node = null;
        this.last = node;
        this.first = node;
        this.size = 0;
        ++this.modCount;
    }
    
    @Override
    public E get(int n) {
        this.checkElementIndex(n);
        return this.node(n).item;
    }
    
    @Override
    public E set(int n, E item) {
        this.checkElementIndex(n);
        LinkedList.Node<E> node = this.node(n);
        Object item2 = node.item;
        node.item = item;
        return item2;
    }
    
    @Override
    public void add(int n, E e) {
        this.checkPositionIndex(n);
        if (n == this.size) {
            this.linkLast(e);
        }
        else {
            this.linkBefore(e, this.node(n));
        }
    }
    
    @Override
    public E remove(int n) {
        this.checkElementIndex(n);
        return this.unlink(this.node(n));
    }
    
    private boolean isElementIndex(int n) {
        return n >= 0 && n < this.size;
    }
    
    private boolean isPositionIndex(int n) {
        return n >= 0 && n <= this.size;
    }
    
    private String outOfBoundsMsg(int i) {
        return "Index: " + i + ", Size: " + this.size;
    }
    
    private void checkElementIndex(int n) {
        if (!this.isElementIndex(n)) {
            throw new IndexOutOfBoundsException(this.outOfBoundsMsg(n));
        }
        return;
    }
    
    private void checkPositionIndex(int n) {
        if (!this.isPositionIndex(n)) {
            throw new IndexOutOfBoundsException(this.outOfBoundsMsg(n));
        }
        return;
    }
    
    LinkedList.Node<E> node(int n) {
        if (n < this.size >> 1) {
            LinkedList.Node<Object> node2 = this.first;
            for (int j = 0; j < n; ++j) {
                node2 = node2.next;
            }
            return node2;
        }
        LinkedList.Node<Object> node = this.last;
        for (int i = this.size - 1; i > n; --i) {
            node = node.prev;
        }
        return node;
    }
    
    @Override
    public int indexOf(Object o) {
        int n = 0;
        if (o == null) {
            LinkedList.Node<Object> node2 = this.first;
            while (node2 != null) {
                if (node2.item == null) {
                    return n;
                }
                ++n;
                node2 = node2.next;
            }
        }
        else {
            LinkedList.Node<Object> node = this.first;
            while (node != null) {
                if (o.equals(node.item)) {
                    return n;
                }
                ++n;
                node = node.next;
            }
        }
        return -1;
    }
    
    @Override
    public int lastIndexOf(Object o) {
        int size = this.size;
        if (o == null) {
            LinkedList.Node<Object> node2 = this.last;
            while (node2 != null) {
                --size;
                if (node2.item == null) {
                    return size;
                }
                node2 = node2.prev;
            }
        }
        else {
            LinkedList.Node<Object> node = this.last;
            while (node != null) {
                --size;
                if (o.equals(node.item)) {
                    return size;
                }
                node = node.prev;
            }
        }
        return -1;
    }
    
    @Override
    public E peek() {
        LinkedList.Node<Object> first = this.first;
        return (first == null) ? null : first.item;
    }
    
    @Override
    public E element() {
        return this.getFirst();
    }
    
    @Override
    public E poll() {
        LinkedList.Node<Object> first = this.first;
        return (first == null) ? null : this.unlinkFirst(first);
    }
    
    @Override
    public E remove() {
        return this.removeFirst();
    }
    
    @Override
    public boolean offer(E e) {
        return this.add(e);
    }
    
    @Override
    public boolean offerFirst(E e) {
        this.addFirst(e);
        return true;
    }
    
    @Override
    public boolean offerLast(E e) {
        this.addLast(e);
        return true;
    }
    
    @Override
    public E peekFirst() {
        LinkedList.Node<Object> first = this.first;
        return (first == null) ? null : first.item;
    }
    
    @Override
    public E peekLast() {
        LinkedList.Node<Object> last = this.last;
        return (last == null) ? null : last.item;
    }
    
    @Override
    public E pollFirst() {
        LinkedList.Node<Object> first = this.first;
        return (first == null) ? null : this.unlinkFirst(first);
    }
    
    @Override
    public E pollLast() {
        LinkedList.Node<Object> last = this.last;
        return (last == null) ? null : this.unlinkLast(last);
    }
    
    @Override
    public void push(E e) {
        this.addFirst(e);
    }
    
    @Override
    public E pop() {
        return this.removeFirst();
    }
    
    @Override
    public boolean removeFirstOccurrence(Object o) {
        return this.remove(o);
    }
    
    @Override
    public boolean removeLastOccurrence(Object o) {
        if (o == null) {
            LinkedList.Node<E> node2 = this.last;
            while (node2 != null) {
                if (node2.item == null) {
                    this.unlink(node2);
                    return true;
                }
                node2 = node2.prev;
            }
        }
        else {
            LinkedList.Node<E> node = this.last;
            while (node != null) {
                if (o.equals(node.item)) {
                    this.unlink(node);
                    return true;
                }
                node = node.prev;
            }
        }
        return false;
    }
    
    @Override
    public ListIterator<E> listIterator(int n) {
        this.checkPositionIndex(n);
        return new LinkedList.ListItr(n);
    }
    
    @Override
    public Iterator<E> descendingIterator() {
        return new LinkedList.DescendingIterator();
    }
    
    private LinkedList<E> superClone() {
        try {
            return (LinkedList)super.clone();
        }
        catch (CloneNotSupportedException cause) {
            throw new InternalError(cause);
        }
    }
    
    public Object clone() {
        LinkedList<E> superClone;
        LinkedList<LinkedList.Node> list = superClone = this.superClone();
        LinkedList.Node<E> node = null;
        list.last = node;
        superClone.first = node;
        list.size = 0;
        list.modCount = 0;
        for (LinkedList.Node node2 = this.first; node2 != null; node2 = node2.next) {
            list.add(node2.item);
        }
        return list;
    }
    
    @Override
    public Object[] toArray() {
        Object[] array = new Object[this.size];
        int n = 0;
        for (LinkedList.Node node = this.first; node != null; node = node.next) {
            array[n++] = node.item;
        }
        return array;
    }
    
    @Override
    public <T> T[] toArray(T[] array) {
        if (array.length < this.size) {
            array = (Object[])(Object[])Array.newInstance(array.getClass().getComponentType(), this.size);
        }
        int n = 0;
        for (LinkedList.Node node = this.first; node != null; node = node.next) {
            array[n++] = node.item;
        }
        if (array.length > this.size) {
            array[this.size] = null;
            return array;
        }
        return array;
    }
    
    private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
        objectOutputStream.defaultWriteObject();
        objectOutputStream.writeInt(this.size);
        for (LinkedList.Node node = this.first; node != null; node = node.next) {
            objectOutputStream.writeObject(node.item);
        }
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        objectInputStream.defaultReadObject();
        for (int int1 = objectInputStream.readInt(), i = 0; i < int1; ++i) {
            this.linkLast(objectInputStream.readObject());
        }
    }
    
    @Override
    public Spliterator<E> spliterator() {
        return new LinkedList.LLSpliterator(this, -1, 0);
    }
    
    private class ListItr implements ListIterator<E>
    {
        private LinkedList.Node<E> lastReturned;
        private LinkedList.Node<E> next;
        private int nextIndex;
        private int expectedModCount;
        
        ListItr(int nextIndex) {
            this.expectedModCount = LinkedList.this.modCount;
            this.next = ((nextIndex == LinkedList.this.size) ? null : LinkedList.this.node(nextIndex));
            this.nextIndex = nextIndex;
        }
        
        @Override
        public boolean hasNext() {
            return this.nextIndex < LinkedList.this.size;
        }
        
        @Override
        public E next() {
            this.checkForComodification();
            if (!this.hasNext()) {
                throw new NoSuchElementException();
            }
            this.lastReturned = this.next;
            this.next = this.next.next;
            ++this.nextIndex;
            return this.lastReturned.item;
        }
        
        @Override
        public boolean hasPrevious() {
            return this.nextIndex > 0;
        }
        
        @Override
        public E previous() {
            this.checkForComodification();
            if (!this.hasPrevious()) {
                throw new NoSuchElementException();
            }
            LinkedList.Node<E> node = (this.next == null) ? LinkedList.this.last : this.next.prev;
            this.next = node;
            this.lastReturned = node;
            --this.nextIndex;
            return this.lastReturned.item;
        }
        
        @Override
        public int nextIndex() {
            return this.nextIndex;
        }
        
        @Override
        public int previousIndex() {
            return this.nextIndex - 1;
        }
        
        @Override
        public void remove() {
            this.checkForComodification();
            if (this.lastReturned == null) {
                throw new IllegalStateException();
            }
            LinkedList.Node<Object> next = this.lastReturned.next;
            LinkedList.this.unlink(this.lastReturned);
            if (this.next == this.lastReturned) {
                this.next = next;
            }
            else {
                --this.nextIndex;
            }
            this.lastReturned = null;
            ++this.expectedModCount;
            return;
        }
        
        @Override
        public void set(E item) {
            if (this.lastReturned == null) {
                throw new IllegalStateException();
            }
            this.checkForComodification();
            this.lastReturned.item = item;
            return;
        }
        
        @Override
        public void add(E e) {
            this.checkForComodification();
            this.lastReturned = null;
            if (this.next == null) {
                LinkedList.this.linkLast(e);
            }
            else {
                LinkedList.this.linkBefore(e, this.next);
            }
            ++this.nextIndex;
            ++this.expectedModCount;
        }
        
        @Override
        public void forEachRemaining(Consumer<? super E> consumer) {
            Objects.requireNonNull(consumer);
            while (LinkedList.this.modCount == this.expectedModCount && this.nextIndex < LinkedList.this.size) {
                consumer.accept(this.next.item);
                this.lastReturned = this.next;
                this.next = this.next.next;
                ++this.nextIndex;
            }
            this.checkForComodification();
        }
        
        final void checkForComodification() {
            if (LinkedList.this.modCount != this.expectedModCount) {
                throw new ConcurrentModificationException();
            }
            return;
        }
    }
    
    private static class Node<E>
    {
        E item;
        LinkedList.Node<E> next;
        LinkedList.Node<E> prev;
        
        Node(LinkedList.Node<E> prev, E item, LinkedList.Node<E> next) {
            this.item = item;
            this.next = next;
            this.prev = prev;
        }
    }
    
    private class DescendingIterator implements Iterator<E>
    {
        private final LinkedList.ListItr itr;
        
        private DescendingIterator() {
            this.itr = new LinkedList.ListItr(LinkedList.this.size());
        }
        
        @Override
        public boolean hasNext() {
            return this.itr.hasPrevious();
        }
        
        @Override
        public E next() {
            return this.itr.previous();
        }
        
        @Override
        public void remove() {
            this.itr.remove();
        }
    }
    
    static final class LLSpliterator<E> implements Spliterator<E>
    {
        static final int BATCH_UNIT = 1024;
        static final int MAX_BATCH = 33554432;
        final LinkedList<E> list;
        LinkedList.Node<E> current;
        int est;
        int expectedModCount;
        int batch;
        
        LLSpliterator(LinkedList<E> list, int est, int expectedModCount) {
            this.list = list;
            this.est = est;
            this.expectedModCount = expectedModCount;
        }
        
        final int getEst() {
            int est;
            if ((est = this.est) < 0) {
                LinkedList<Object> list;
                if ((list = this.list) == null) {
                    int est2 = 0;
                    this.est = est2;
                    est = est2;
                }
                else {
                    this.expectedModCount = list.modCount;
                    this.current = list.first;
                    int size = list.size;
                    this.est = size;
                    est = size;
                }
            }
            return est;
        }
        
        @Override
        public long estimateSize() {
            return (long)this.getEst();
        }
        
        @Override
        public Spliterator<E> trySplit() {
            int est = this.getEst();
            LinkedList.Node<Object> current;
            if (est > 1 && (current = this.current) != null) {
                int n = this.batch + 1024;
                if (n > est) {
                    n = est;
                }
                if (n > 33554432) {
                    n = 33554432;
                }
                Object[] array = new Object[n];
                int batch = 0;
                do {
                    array[batch++] = current.item;
                } while ((current = current.next) != null && batch < n);
                this.current = current;
                this.batch = batch;
                this.est = est - batch;
                return Spliterators.spliterator(array, 0, batch, 16);
            }
            else {
                return null;
            }
        }
        
        @Override
        public void forEachRemaining(Consumer<? super E> consumer) {
            if (consumer == null) {
                throw new NullPointerException();
            }
            int est;
            LinkedList.Node<Object> node;
            if ((est = this.getEst()) > 0 && (node = this.current) != null) {
                this.current = null;
                this.est = 0;
                do {
                    Object item = node.item;
                    node = node.next;
                    consumer.accept(item);
                } while (node != null && --est > 0);
            }
            if (this.list.modCount != this.expectedModCount) {
                throw new ConcurrentModificationException();
            }
            return;
        }
        
        @Override
        public boolean tryAdvance(Consumer<? super E> consumer) {
            if (consumer == null) {
                throw new NullPointerException();
            }
            LinkedList.Node<Object> current;
            if (this.getEst() > 0 && (current = this.current) != null) {
                --this.est;
                Object item = current.item;
                this.current = current.next;
                consumer.accept(item);
                if (this.list.modCount != this.expectedModCount) {
                    throw new ConcurrentModificationException();
                }
                return true;
            }
            else {
                return false;
            }
        }
        
        @Override
        public int characteristics() {
            return 16464;
        }
    }
}
